// Primitives.cc   New primitives not in primitives.cxx
// W. Langdon cs.ucl.ac.uk 23 January 1995
//                     based upon queue2.cc 20 Dec 1994 Revision: 1.57
//                     and primitiv.cxx from Andy Singleton
// $Revision: 1.47 $

//#define DEBUG1 true

//Modifications (in reverse order)

//WBL 15 Apr 97  Add ORIGINAL_LOOP for simple inner_loop_value behaviour

//WBL 12 Apr 97  Ensure LIST_LIB does not get (address1==0) bugfix

//WBL  9 Apr 97  Ensure calleval arg2 != -2 since code is nasty frig which
//               may not work on Alpha

//WBL 27 May 96  BUGFIX ensure i0 is saved on entering loop not at start of
//               forxxx primitive. ie limits use outer scope not inner

//WBL  5 May 96  Add R0..R9

//WBL 17 Apr 96  Add IfltEval

//WBL  7 Apr 96  BUGFIX ensure loop summing EVAL answers includes last value

//WBL  4 Mar 96  Add ForSumEval and change loop to accomodate it
//               Add ForMinSumEval
//               Add i1 to i4

//WBL 26 Oct 95  Ensure modEval uses int in calculations rather than retval
//               Add intEval, setEval

//WBL 15 Oct 95  Add calling opt_time in ifeqeval

//WBL 28 Sep 95  Add NegoneEval

//WBL 22 Sep 95  Replace list.h by prob.h

//WBL  2 Sep 95  BUGFIX Add support for VAR_TIMINGS

//WBL  1 Sep 95  BUGFIX Ensure inner_loop_value is always zero outside loops

//WBL 30 Aug 95  Support nested loops, add i0_not_in_loop

//WBL 29 Aug 95  Protect swapEval from itself overwritting contents of
//               buffer word store[0]. 

//WBL 23 Jun 95  Allow partial matching on string length in print_match

//WBL 13 Jun 95  Make comptaible with gcc version 2.6.3

//WBL 12 May 95  Add incrementing Memory_errors on memory addressing error

//WBL  1 May 95  Add TRACE_RUN

//WBL 26 Apr 95  BUGFIX print_match says complete string match if all 
//               components are there but they are in the wrong order, fixit

//WBL 23 Apr 95  Allow aux1 as cache input if no arguments to function

//WBL 22 Apr 95  Ensure ADF_CACHE still works, Move def to Primitives.h

//WBL 22 Apr 95  BUGFIX. Correct value returned by forwhile loop on EVAL=0

//WBL 20 Apr 95  Add TIMINGS

//WBL 18 Apr 95  BUGFIX. Correct location of IP on exiting forwhile loop
 
//WBL 18 Apr 95  Add swapEval

//WBL  8 Apr 95  Make arg2_address int rather than bool

//WBL 31 Mar 95  Add ARG1Eval and ARG2EVAL, AddF5 etc

//WBL 31 Mar 95  Add forwhileEval, AddF7, FUNCEval
//               add update_arg1 to calleval

//WBL 11 Mar 95  Add tree_contains and sub_tree_contains

//WBL 21 Feb 95  Performance changes: inline get_address. Make loop remember
//               where end of loop is.

//WBL 19 Feb 95  Add nested_loop
//               Allow loops withing end point calculations of from and down

//WBL 16 Feb 95  Allow partial matching on print buffer

//WBL  2 Feb 95  In while loop, Start I0 from 0 rather than 1

//WBL 23 Jan 95  New file

#include <assert.h>
#include "pch.h"
#include "chrome.h"
#include "prob.h"
#ifdef TIMINGS
#include "time.h"
#endif
#ifdef TRACE_RUN
#include "trace.h"
#endif
#include "Primitives.h"
#include "gp.h"
#ifdef STACK_PRIM
#include "stack_prim.h"
#endif
extern char scratch[100];
#ifdef FASTEVAL
	extern evalnode* IpGlobal;
	extern Chrome* ChromeGlobal;
#endif

#define Max               10

retval Arg1, Arg2;
retval ARG1, ARG2;
retval Aux1, Aux2, Aux3, store[store_limit];
int store_used[store_limit];

#ifdef BYREF
#define Vnone (-4)
#define Vaux3 (-3)
#define Vaux2 (-2)
#define Vaux1 (-1)
//else store[x]
int last_var_used = Vnone;
#endif //BYREF

int Primitive_error;
int Memory_errors;

#ifdef ADF
int calls [ NUM_TREES]; //used to prevent recursive calls
#endif //ADF

#ifdef ORIGINAL_LOOP
retval inner_loop_value;
#else /*ORIGINAL_LOOP*/
retval loop_value[loop_max_depth+2];
#if(loop_max_depth==1)
retval inner_loop_value;
#endif
#endif /*ORIGINAL_LOOP*/
int iterate_count;

int loop_depth;


#ifdef GEN_MEM
void myfitnessvalue::load_genetic_memory() 
{
	memcpy(genetic_memory,
	       &store[1+(-Array_bot)],
	       GENETIC_MEM_SIZE*sizeof(retval));
}

void myfitnessvalue::read_genetic_memory() 
{
	memcpy(&store[1+(-Array_bot)],
	       genetic_memory,
	       GENETIC_MEM_SIZE*sizeof(retval));
}
#endif

int get_address()
{
int address = int((EVAL)+(-Array_bot)+1);

if ( address <= 0 || address >= store_limit )
  {
	Primitive_error |= memory_error;
	Memory_errors++;
	store[0] = 0;        //Prevent using buffer area (depends on write)
	return 0;             //safe buffer area
  }
else
  {	return address;
  }
}//end get_address
//**************************************************************************

#ifdef ADF_CACHE
void retval_cache::stats_init()
{
	calls = 0;
	hits  = 0;
	magic_hits = 0;
	min   = MAXINT;
	max   = -MAXINT;
}

//retval_cache adf1_cache(1000);
#endif //ADF_CACHE

//**************************************************************************
/////////////////// Problem specific functions and terminals
// Use the OPDEF,EVAL,IP and CHROMEP defines
// and the code will recompile for different eval methods

// Problem Specific Terminals

// gcc refuses to accept these in primitiv So stuff em here!

OPDEF(Prog2Eval) {EVAL; return EVAL;}
//OPDEF(Prog3Eval) {EVAL;EVAL; return EVAL;}
OPDEF(Qrog2Eval) {
        retval first_eval=EVAL;
#ifdef BYREF
        int    save_last_var_used = last_var_used;
#endif //BYREF
	EVAL;
#ifdef BYREF
	last_var_used = save_last_var_used;
#endif //BYREF
	return first_eval;}
//OPDEF(Qrog3Eval) {retval first_eval=EVAL;EVAL;EVAL; return first_eval;}

OPDEF(zeroEval) {return 0;}
OPDEF(oneEval) {return 1;}
OPDEF(NegoneEval) {return -1;}
OPDEF(twoEval) {return 2;}
OPDEF(MaxEval) {return Max;}
OPDEF(BotEval) {return Array_bot;}
OPDEF(TopEval) {return Array_top;}
OPDEF(R0Eval) {return RandomValue[0];}
OPDEF(R1Eval) {return RandomValue[1];}
OPDEF(R2Eval) {return RandomValue[2];}
OPDEF(R3Eval) {return RandomValue[3];}
OPDEF(R4Eval) {return RandomValue[4];}
OPDEF(R5Eval) {return RandomValue[5];}
OPDEF(R6Eval) {return RandomValue[6];}
OPDEF(R7Eval) {return RandomValue[7];}
OPDEF(R8Eval) {return RandomValue[8];}
OPDEF(R9Eval) {return RandomValue[9];}

OPDEF(modEval) { //see workbook III 10 June 1994
                 //    workbook IV  3 August 1994
       int a = int(EVAL);
       int b = abs(int(EVAL));
       if (b!=0) return retval((a % b + b) % b);
       else      return retval(a);
}
OPDEF(intEval) {return int(EVAL);}
OPDEF(Arg1Eval) {return Arg1;}
OPDEF(Arg2Eval) {return Arg2;}
OPDEF(ARG1Eval) {return ARG1;}
OPDEF(ARG2Eval) {return ARG2;}
OPDEF(Aux1Eval) {
#ifdef BYREF
        last_var_used=Vaux1;
#endif //BYREF
        return Aux1;}
OPDEF(Aux2Eval) {
#ifdef BYREF
        last_var_used=Vaux2;
#endif //BYREF
        return Aux2;}
OPDEF(Aux3Eval) {
#ifdef BYREF
        last_var_used=Vaux3;
#endif //BYREF
        return Aux3;}
OPDEF(inc_Aux1Eval)  {
#ifdef BYREF
        last_var_used=Vaux1; 
#endif //BYREF
        return ++Aux1;}
OPDEF(Minc_Aux1Eval) {
#ifdef BYREF
        last_var_used=Vaux1; 
#endif //BYREF
        return Aux1 = ((int(++Aux1))%Max+Max)%Max;}
OPDEF(dec_Aux1Eval)  {
#ifdef BYREF
        last_var_used=Vaux1;
#endif //BYREF
        return --Aux1;}
OPDEF(inc_Aux2Eval)  {
#ifdef BYREF
        last_var_used=Vaux2; 
#endif //BYREF
        return ++Aux2;}
OPDEF(Minc_Aux2Eval) {
#ifdef BYREF
        last_var_used=Vaux2;
#endif //BYREF
        return Aux2 = ((int(++Aux2))%Max+Max)%Max;}
OPDEF(dec_Aux2Eval)  {
#ifdef BYREF
        last_var_used=Vaux2;
#endif //BYREF
        return --Aux2;}
OPDEF(inc_Aux3Eval)  {
#ifdef BYREF
        last_var_used=Vaux3;
#endif //BYREF
        return ++Aux3;}
OPDEF(Minc_Aux3Eval) {
#ifdef BYREF
        last_var_used=Vaux3;
#endif //BYREF
        return Aux3 = ((int(++Aux3))%Max+Max)%Max;}
OPDEF(dec_Aux3Eval)  {
#ifdef BYREF
        last_var_used=Vaux3;
#endif //BYREF
        return --Aux3;}

retval readdata(int addr)
{
int address = addr+(-Array_bot)+1;

if ( address <= 0 || address >= store_limit )
  {
	Primitive_error |= memory_error;
	Memory_errors++;
	return 0;
  }
else {
       if (store_used[address]==0)
               store_used[address] = store_read_unassigned;
#ifdef BYREF
       last_var_used = address;
#endif //BYREF
       return store[address];
}
}//end readdata

OPDEF(readEval) {
       int address = get_address();
       if (store_used[address]==0)
               store_used[address] = store_read_unassigned;
#ifdef BYREF
       last_var_used = address;
#endif //BYREF
       return store[address];
}
OPDEF(writeEval) {
       int address = get_address(); 
       retval old = store[address];
       store[address] = EVAL;
       store_used[address] = store_used[address] | store_written;
#ifdef BYREF
       last_var_used = address;
#endif //BYREF
       return old;                // copy what Tesler does
}
OPDEF(setEval) {
       int address = get_address(); 
       store[address] = EVAL;
       store_used[address] = store_used[address] | store_written;
#ifdef BYREF
       last_var_used = address;
#endif //BYREF
       return store[address];
}
OPDEF(swapEval) {
       int address1 = get_address();
       if (store_used[address1]==0)
               store_used[address1] = store_read_unassigned;
       int address2 = get_address();
       if (store_used[address2]==0)
               store_used[address2] = store_read_unassigned;
       store_used[address1] |= store_written;
       store_used[address2] |= store_written;
#ifdef LIST_LIB
       retval temp = store[address1];
#else
       retval temp = (address1==0)? 0:store[address1];
#endif
       store[address1] = store[address2];
       store[address2] = temp;
#ifdef BYREF
       last_var_used = address2;
#endif //BYREF
       return 1;
}
void blockset(const int address, const int size, const retval data[])
{
	if(address<=0 || (address+size-1) >= store_limit ) return;
	for(int i = address; i<=(address+size-1); i++) {
		store_used[i] |= store_written;}
#ifdef BYREF
	last_var_used = address+size-1;
#endif //BYREF
	memcpy(&store[address],data,size*sizeof(retval));
	return;
}//end blockset

//OPDEF(write_Aux1Eval) {
//       retval old = Aux1;
//       Aux1 = EVAL;
//       return old;                // copy what write (ie Tesler) does
//}
//OPDEF(write_Aux2Eval) {
//       retval old = Aux2;
//       Aux2 = EVAL;
//       return old;                // copy what write (ie Tesler) does
//}
OPDEF(set_Aux1Eval) {
       Aux1 = EVAL;
#ifdef BYREF
       last_var_used = Vaux1;
#endif //BYREF
       return Aux1;                // Nb simplier than Tesler
}
OPDEF(set_Aux2Eval) {
       Aux2 = EVAL;
#ifdef BYREF
       last_var_used = Vaux2;
#endif //BYREF
       return Aux2;                // Nb simplier than Tesler
}
OPDEF(set_Aux3Eval) {
       Aux3 = EVAL;
#ifdef BYREF
       last_var_used = Vaux3;
#endif //BYREF
       return Aux3;                // Nb simplier than Tesler
}

#ifdef ADF
#ifdef ADF_CACHE
retval calleval (int numarg, int tree, 
                 BOOL update_arg1, int arg2_address,
                 retval_cache* cache)
#else
#ifdef BYREF
retval calleval (int numarg, int tree, 
                 BOOL update_arg1/* = FALSE*/, int arg2_address /* = -1*/
//		 ,evalnode* target = NULL
		 )
#else
retval calleval (int numarg, int tree )
#endif //BYREF
#endif //ADF_CACHE
{
#ifdef TIMINGS
	ThisTimmer->base_time(tree);
#endif
#ifdef TRACE_RUN
	ThisTrace->trace(tree);
#endif
	retval result;

	retval save_arg1 = Arg1;
	retval save_arg2 = Arg2;
#ifdef BYREF
	int save_var_used = last_var_used;
#endif //BYREF

	retval t1 = 0;
	retval t2 = 0;
#ifdef BYREF
	int var_used = Vnone;
#endif //BYREF

	if ( numarg > 0 )
		{ 
#ifdef BYREF
                  last_var_used = Vnone; 
#endif //BYREF
		  t1 = EVAL;
#ifdef BYREF
		  var_used = last_var_used;
#endif //BYREF
	        }
	if ( numarg > 1 ) {
#ifdef ADF_CACHE 
#define ADF_OR_BYREF true
#endif

#ifdef BYREF
#define ADF_OR_BYREF true
#endif

#ifdef ADF_OR_BYREF
		if     (arg2_address == -1) 
			t2 = EVAL;
		else if(arg2_address == -2) {
assert(0==-2);		//t2 = IP;
			IP++;           // Jump the expression
			TRAVERSE();
			IP--;           // back up one (since EVAL increments)
		}
		else
			t2 = arg2_address;
#else
		t2 = EVAL;
#endif
	}
	assert ( numarg <= 2 ); 

	Arg1 = t1;
	Arg2 = t2;

//	if ( (target == NULL) && (calls [ tree ]++ > 1) )
	if ( calls [ tree ]++ > 1 )
	      { Primitive_error |= recurse_error;
	        result = 0;
	      }
	else  { 
#ifdef DEBUG1
		cout<<"Tree ("<<tree<<") ";
		if(numarg==1) cout<<"Arg1="<<Arg1<<" "<<flush;
		if(numarg==0) cout<<"Aux1="<<Aux1<<" "<<flush;
#endif //DEBUG1
#ifdef ADF_CACHE
		if (cache==NULL || 
#ifdef VAR_TIMINGS
		    (numarg==1 &&cache->held(Arg1,&result,ThisTimmer)==FALSE)||
		    (numarg==0 &&cache->held(Aux1,&result,ThisTimmer)==FALSE))
#else
		    (numarg==1 && cache->held(Arg1,&result)==FALSE) ||
		    (numarg==0 && cache->held(Aux1,&result)==FALSE)   )
#endif
#endif
		{
#ifdef ADF_CACHE
#ifdef VAR_TIMINGS
			run_time::timetyp start_time = ThisTimmer->timmer();
			int start_loop = iterate_count;
#endif
#endif
#ifdef ORIGINAL_LOOP
			retval save_inner_loop_value = inner_loop_value;
			inner_loop_value = 0;
#else
			retval save_loop_value[loop_max_depth+2];
			memcpy(save_loop_value,loop_value,sizeof(loop_value));
			memset(loop_value,0,sizeof(loop_value));
#endif /*ORIGINAL_LOOP*/
			evalnode* save_ip = IP;
//			if(target==NULL)
				result = CHROMEP->evalAll ( tree );
#ifdef DEBUG1
			cout<<"evalutes to "<<flush;
#endif //DEBUG1
//			else {
//				IP = target;
//				result = EVAL;
//			}
			IP = save_ip;
#ifdef ORIGINAL_LOOP
			inner_loop_value = save_inner_loop_value;
#else
			memcpy(loop_value,save_loop_value,sizeof(loop_value));
#endif /*ORIGINAL_LOOP*/
#ifdef ADF_CACHE
#ifdef VAR_TIMINGS
			if ( (cache!=NULL) &&
			    ((Primitive_error & loop_too_long_error) == 0)){
				if(numarg==0) cache->add(Aux1,result,
					       ThisTimmer->timmer()-start_time,
					       iterate_count-start_loop);
				else          cache->add(Arg1,result,
					       ThisTimmer->timmer()-start_time,
					       iterate_count-start_loop);}
#else
			if (cache!=NULL)
				if(numarg==0) cache->add(Aux1,result);
				else          cache->add(Arg1,result);
#endif
#endif
	        }
#ifdef DEBUG1
                cout<<result<<endl;
#endif //DEBUG1
#ifdef BYREF
		if(update_arg1) {
//cout<<"calleval updating "<<last_var_used<<" to "<<result<<endl;
		switch(last_var_used)
			{
			case Vnone:
				break;
			case Vaux1:
				Aux1 = result;
				break;
			case Vaux2:
				Aux2 = result;
				break;
			case Vaux3:
				Aux3 = result;
				break;
			default:
				if ( last_var_used > 0 &&
				     last_var_used < store_limit)
				    {
				    store[last_var_used]       = result;
				    store_used[last_var_used] |= store_written;
				    }
				break;
			}//end switch
			}//end update_arg1
#endif //BYREF
	      }
//	if (target == NULL) calls [ tree ]--;
	calls [ tree ]--;

	Arg1   = save_arg1;
	Arg2   = save_arg2;
#ifdef BYREF
	last_var_used = save_var_used;
#endif //BYREF

	return result;

}//end calleval

#ifdef MINC_ADF
int static_check_loop = 0; //detect loops

int static_check(Chrome* chrome, int tree, int& values)
// Perform static analysis of new chromsome. Just changed in tree
// Crude but should demonstrate if the idea is worth pursuing
{

const int good = 0;
const int recurse_penalty = 100;
const int poor_function_penalty = 100;
//const int memory_penalty = 20;

//store_written - not used directly by static_check

static_check_loop++;
if(static_check_loop > 1000)
{   cout<<"static_check looping! ("<<static_check_loop<<")"<<endl;
    static_check_loop = 0;
    chrome->write(PRETTY_NONE,cout); cout<<endl;
    return good;               //stop it looping
}

const int num_checks = 2*(Max + 1);

//cout<<"static_check (ADF1) "<<static_check_loop<<flush; //debug

retval ans [num_checks];

BOOL follow    = FALSE;
BOOL new_value = FALSE;
int  k         = -1;
values         = 0; //assume the worst
for (int i = 0; (k<Max) && (i<num_checks); i++)
       {//start call adf1 loop
	 int input;
	 if (!follow && new_value && (ans[i-1] < 0 || ans[i-1] > Max) )
         {  input = ans[i-1];
	    follow = TRUE;
         }
	 else
         {  input  = ++k;
	    follow = FALSE;
         }

	 //initialise everything to be consistent with queue::fitness

	 Primitive_error = 0;
	 Memory_errors = 0;
	 Aux1 = 0;
	 Aux2 = 0;
	 Aux3 = 0;
	 memset(store,0,sizeof(store));?

	 Arg1 = input;
	 Arg2 = 0;
	 memset(calls,0,sizeof(calls));
	 calls [ tree ] = 1; //should be in evalAll, but avoid chrome.cxx

	 chrome->SetupEval();
	 ans [i] = chrome->evalAll ( tree );

//if(recurse_error)cout<<" RECURSE_ERROR";
//if(memory_error) cout<<" MEMORY_ERROR";
//cout<<" "<<Arg1<<"=>"<<ans[i]<<" "<<flush;

//cout<<" "<<last_var_used<<" "<<arg1_used<<" "<<arg1_side_effect<<" ";

	 if ((Primitive_error|recurse_error)!=0)  return recurse_penalty;

         new_value = TRUE;
	 for (int j = 0; j < i; j++)
	 {
		 if (ans[i] == ans[j]) new_value = FALSE;
	 }
	 if (new_value) values++;

	if (ans[i] == i) return poor_function_penalty; //identity
        };//end run adf1 to test it loop


if (values <= 1)
	return poor_function_penalty; //ie its a constant
else
  {
//chrome->write(PRETTY_NONE,cout); cout<<endl; //debug
        static_check_loop = 0;
	return good;
  }
}//end static_check
#endif //MINC_ADF
#endif //ADF


OPDEF(IfeqEval)	//Ifeq(condition1,condition2,dothis,dothat)
{
	retval rval;
        rval=EVAL;
	if (rval==EVAL)
	{
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
		IP++;           // Jump the third expression
		TRAVERSE();
		IP--;           // And back up one (since EVAL increments)
	}
	else {
		IP++;
		TRAVERSE();    // Jump the second expression
		IP--;                                   // Back up for EVAL
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
	}
	return rval;
}

OPDEF(IfltEval)	//Iflt(condition1,condition2,dothis,dothat)
{
	retval rval;
        rval=EVAL;
	if (rval<EVAL)
	{
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
		IP++;           // Jump the third expression
		TRAVERSE();
		IP--;           // And back up one (since EVAL increments)
	}
	else {
		IP++;
		TRAVERSE();    // Jump the second expression
		IP--;                                   // Back up for EVAL
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
	}
	return rval;
}//end IfltEval


retval loop(int step)
{
retval rval = 0;
if (loop_depth>=loop_max_depth) {
#ifdef DEBUG
	cout<<"Loop too nested "<<IP<<flush;
#endif
	Primitive_error |= loop_too_nested_error;
	IP++;
	if(step==0) { //skip two sub trees
		TRAVERSE();
		TRAVERSE();
	}
	else {        //skip three sub trees
		TRAVERSE();
		TRAVERSE();
		TRAVERSE();
	}
	IP--;
#ifdef DEBUG
	cout<<" skipped to "<<IP<<endl;
#endif
}
else {
	evalnode* end_ip = NULL;
#ifdef ORIGINAL_LOOP
	retval save_inner_loop_value = inner_loop_value;
#endif /*ORIGINAL_LOOP*/
	if(step==0) {
		evalnode* start_ip = IP;
		retval dummy1;
#ifndef ORIGINAL_LOOP
		memmove(&loop_value[1],&loop_value[0],loop_depth*sizeof(retval));
#endif
		loop_depth++;
		inner_loop_value = 0;
		while((0!=(dummy1 = EVAL))&&(iterate_count<=max_iterates)) {
#ifdef TIMINGS
			ThisTimmer->opt_time(IP);
#endif
			retval dummy2 = EVAL;
			end_ip = IP;
			inner_loop_value++;
			iterate_count++;
			IP = start_ip;
		}
	}
	else {  
		int overlap;
		retval new_inner_loop_value;
		if(step==3) {
			new_inner_loop_value = 1;
			overlap = int(EVAL);
			if (overlap <= 0) overlap = 1;
		}
		else {
			new_inner_loop_value = EVAL;
		}
		retval limit=EVAL;
		evalnode* start_ip = IP;
#ifndef ORIGINAL_LOOP
		memmove(&loop_value[1],&loop_value[0],loop_depth*sizeof(retval));
#endif
		loop_depth++;
		inner_loop_value = new_inner_loop_value;
		switch (step) {
		case -1 : for(; (inner_loop_value >= limit)        &&
			           (iterate_count <= max_iterates) ;
			     iterate_count++,IP = start_ip) {
#ifdef TIMINGS
			          ThisTimmer->opt_time(IP);
#endif
			          retval dummy = EVAL;
			          end_ip = IP;
			          inner_loop_value--;
		         }
		         break;

		case  1 : for(; (inner_loop_value <= limit)        && 
			           (iterate_count <= max_iterates) ;
			     iterate_count++,IP = start_ip) {
#ifdef TIMINGS
			          ThisTimmer->opt_time(IP);
#endif
			          retval dummy = EVAL;
			          end_ip = IP;
			          inner_loop_value++;
			  }
			  break;

                case  2 : {BOOL continue_loop = TRUE;
                           for(; (inner_loop_value <= limit)        && 
                                    (iterate_count <= max_iterates) &&
                                           continue_loop            ;
                             iterate_count++,IP = start_ip) {
#ifdef TIMINGS
			          ThisTimmer->opt_time(IP);
#endif
                                  retval dummy = EVAL;
                                  continue_loop = (dummy != 0);
			          if(continue_loop) inner_loop_value++;
			          end_ip = IP;
			  }}
			  break;

		case 3 : { /* ForMinSumEval */
			rval = inner_loop_value;
			int limit2 = ((limit < 1) ||
				      (limit <= max_iterates-iterate_count) )?
			               int(limit): max_iterates-iterate_count;

			if(limit2 < overlap) break; //nopoint running loop 
			                            //answer must be one
//			retval answers[limit2+1];   //incase of rounding error
			retval* answers = new retval [limit2+1];
			for(; (inner_loop_value <= limit)        && 
			         (iterate_count <= max_iterates) ;
			     iterate_count++,IP = start_ip) {
#ifdef TIMINGS
			          ThisTimmer->opt_time(IP);
#endif
			          answers[int(inner_loop_value) - 1] = EVAL;
			          end_ip = IP;
			          inner_loop_value++;
			  }
			retval min_sum = FLT_MAX;
			for(int i=1; i<=(inner_loop_value-overlap); i++)  {
				retval sum = 0;
				for(int j = i; j < i+overlap; j++)
					sum += answers[j-1];
				if(sum < min_sum) {
					min_sum = sum;
					rval = retval(i);
				}
			      
			  }
			delete[] answers;
		      }
			  break;

		case  4 : /* ForSumEval */
			  rval = 0;
			  for(; (inner_loop_value <= limit)        && 
			           (iterate_count <= max_iterates) ;
			     iterate_count++,IP = start_ip) {
#ifdef TIMINGS
			          ThisTimmer->opt_time(IP);
#endif
			          rval += EVAL;
			          end_ip = IP;
			          inner_loop_value++;
			  }
			  break;


		default : assert (0==1);

		     }//end case
	}
	if(end_ip != NULL)
		IP = end_ip;
	else {
		IP++;
		TRAVERSE();    // Jump the loop body
		IP--;          // Back up for EVAL
	}
	if (iterate_count>=max_iterates) 
		Primitive_error |= loop_too_long_error;
	if(step <= 2)
		rval = inner_loop_value;

#ifdef ORIGINAL_LOOP
	inner_loop_value = save_inner_loop_value;
#else
	memmove(&loop_value[0],&loop_value[1],loop_depth*sizeof(retval));
#endif
	loop_depth--;
}
return rval;
}//end loop

BOOL sub_nested_loop(Chrome* chrome, const BOOL in_a_loop, 
		                     const BOOL check_i0_only = FALSE)
{typedef gp* gp_ptr;
	chrome->ip++;
	gp* probl = gp_ptr(chrome->probl);
	int funcnum = FUNCNUM(chrome->expr[chrome->ip]);
	int args=chrome->funclist[funcnum]->argnum;
	if(!in_a_loop && (funcnum==probl->i0_funcnum))        return TRUE;
	if((funcnum==probl->for_funcnum) ||
	   (funcnum==probl->down_funcnum)||
	   (funcnum==probl->forwhile_funcnum)) {
		if(in_a_loop&&!check_i0_only)                 return TRUE;
		for(int i = 0; i<(args-1) ;i++) {
			if(sub_nested_loop(chrome,in_a_loop,check_i0_only))
				                              return TRUE;
		}
		return (sub_nested_loop(chrome,TRUE,check_i0_only));
	}
        else if (funcnum==probl->while_funcnum) { 
		if(in_a_loop&&!check_i0_only)                 return TRUE;
		for(int i = 0; i<args ;i++) {
			if(sub_nested_loop(chrome,TRUE,check_i0_only))
				                              return TRUE;
		}
        } else {
		for(int i = 0;i<args;i++) {
			if(sub_nested_loop(chrome,in_a_loop,check_i0_only))
				                              return TRUE;
	}}

	return FALSE;
}//end sub_nested_loop

BOOL nested_tree_or_i0(Chrome* chrome, int tree)
{
	int save_ip = chrome->ip;
	chrome->ip = chrome->tree_starts[tree] -1;
	BOOL a = sub_nested_loop(chrome,FALSE);
	chrome->ip = save_ip;
	return a;
}//end nested_tree_or_i0

BOOL i0_not_in_loop(Chrome* chrome, int tree)
{
	int save_ip = chrome->ip;
	chrome->ip = chrome->tree_starts[tree] -1;
	BOOL a = sub_nested_loop(chrome,FALSE,TRUE);
	chrome->ip = save_ip;
	return a;
}//end nested_tree_or_i0


OPDEF(ForEval)	//For(start,to,dowork)
{
return loop(1);
}//end OPDEF(ForEval) 

OPDEF(DownEval)	//Down(start,to,dowork)
{
return loop(-1);
}//end OPDEF(ForEval) 

OPDEF(WhileEval) //while(condition<>0,dowork)
{
return loop(0);
}//end OPDEF(ForEval) 

OPDEF(ForWhileEval)	//forwhile(start,to,condition)
{
return loop(2); //For(start, io<=to && condition<>0)
}//end OPDEF(ForWhileEval) 

OPDEF(ForMinSumEval)	//for(overlap,start=1,to,dowork)
{
return loop(3);
}//end OPDEF(ForMinSumEval) 

OPDEF(ForSumEval)	//for(start,to,sum+=dowork)
{
return loop(4);
}//end OPDEF(ForSumEval) 

OPDEF(I0Eval)  {
        return inner_loop_value;}
#ifndef ORIGINAL_LOOP
OPDEF(I1Eval)  {
        return loop_value[1];}
OPDEF(I2Eval)  {
        return loop_value[2];}
OPDEF(I3Eval)  {
        return loop_value[3];}
OPDEF(I4Eval)  {
        return loop_value[4];}
#endif /*ORIGINAL_LOOP*/


// // // // // // // // // // // // // // // // // // // // // // // // 

BOOL sub_tree_contains(Chrome* chrome, int target)
{
	chrome->ip++;
	int funcnum = FUNCNUM(chrome->expr[chrome->ip]);
        if(funcnum==target)                                   return TRUE;
	int args=chrome->funclist[funcnum]->argnum;
	for(int i = 0;i<args;i++) {
		if(sub_tree_contains(chrome,target))          return TRUE;
	}
	return FALSE;
}//end sub_tree_contains

BOOL tree_contains(Chrome* chrome, int tree, int target)
{
	int save_ip = chrome->ip;
	chrome->ip = chrome->tree_starts[tree] -1;
	BOOL a = sub_tree_contains(chrome,target);
	chrome->ip = save_ip;
	return a;

}//end tree_contains


//////////////////////////////////////////////////////////////////////
void indexed_memory::clear( int& seed)
{
	Aux1 = Aux2 = Aux3 = 0;
	if (seed==0) {
		memset(store,     0,sizeof(store));
	}
	else {
		for ( int i = 1; i<store_limit; i++ ) {
			store[i] = random_value(seed);
		}
        }
}

void clear( int& seed)
{
#ifdef ADF
	memset(calls,0,sizeof(calls));
#endif
	Primitive_error = 0;
	Memory_errors = 0;
#ifdef ORIGINAL_LOOP
        inner_loop_value = 0;
#else
        clear_loop_value();
#endif
        iterate_count = 0;

	Arg1 = Arg2 = 0;
	ARG1 = ARG2 = 0;
	Aux1 = Aux2 = Aux3 = 0;
	if (seed==0) {
		memset(store,     0,sizeof(store));
	}
	else {
		for ( int i = 1; i<store_limit; i++ ) {
			store[i] = random_value(seed);
		}
        }
#ifdef STACK_PRIM
	makenullEval();
#endif
}


void indexed_memory::clear_stats()
{
	memset(store_used,0,sizeof(store_used));
}

void clear_stats()
{
	memset(store_used,0,sizeof(store_used));
#ifdef TIMINGS
	ThisTimmer->clear();
#endif
}


void indexed_memory::save_active()
{
	Aux1 = ::Aux1;
	Aux2 = ::Aux2;
	Aux3 = ::Aux3;
	memcpy(store,     ::store,sizeof(store));
	memcpy(store_used,::store_used,sizeof(store_used));
}

void indexed_memory::load_active()
{
	::Aux1 = Aux1;
	::Aux2 = Aux2;
	::Aux3 = Aux3;
	memcpy(::store,     ::store,sizeof(store));
	memcpy(::store_used,::store_used,sizeof(store_used));
}

int indexed_memory::cells_used()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0) used++;}
	return used;
}

int cells_used()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0) used++;}
	return used;
}

int indexed_memory::cells_used_spanned()
{
	int lowest_used  = store_limit + 1;
	int highest_used = -1;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0) {
			if(i>highest_used) highest_used = i;
			if(i<lowest_used ) lowest_used  = i;
		}
	}
	
	return (highest_used<0) ? 0 : highest_used + 1 - lowest_used;
}

int cells_used_spanned()
{
	int lowest_used  = store_limit + 1;
	int highest_used = -1;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0) {
			if(i>highest_used) highest_used = i;
			if(i<lowest_used ) lowest_used  = i;
		}
	}
	
	return (highest_used<0) ? 0 : highest_used + 1 - lowest_used;
}

int indexed_memory::cells_written()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_written)!=0) used++;}
	return used;
}
int cells_written()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_written)!=0) used++;}
	return used;
}

int indexed_memory::cells_written_spanned()
{
	int lowest_used  = store_limit + 1;
	int highest_used = -1;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_written)!=0) {
			if(i>highest_used) highest_used = i;
			if(i<lowest_used ) lowest_used  = i;
		}
	}
	
	return (highest_used<0) ? 0 : highest_used + 1 - lowest_used;
}
int cells_written_spanned()
{
	int lowest_used  = store_limit + 1;
	int highest_used = -1;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_written)!=0) {
			if(i>highest_used) highest_used = i;
			if(i<lowest_used ) lowest_used  = i;
		}
	}
	
	return (highest_used<0) ? 0 : highest_used + 1 - lowest_used;
}


int indexed_memory::cells_read_before_write()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_read_unassigned)!=0) used++;}
	
	return used;
}
int cells_read_before_write()
{
	int used = 0;
	for ( int i = 1; i<store_limit; i++ ) {
		if((store_used[i]&store_read_unassigned)!=0) used++;}
	
	return used;
}


void indexed_memory::write( ostream& fout )
{
	fout<<" Aux1 = " << Aux1;
#ifdef AUX2
	fout<<" Aux2 = " << Aux2;
	fout<<" Aux3 = " << Aux3;
#endif
#ifdef BYREF
	fout<<" lvusd  " <<last_var_used;
#endif //BYREF
	fout<<endl;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0){
			fout <<i + Array_bot - 1<<"\t";
			if((store_used[i]&store_read_unassigned)!=0)
				fout<<"u";
			else
				fout<<" ";
			if((store_used[i]&store_written)!=0)
				fout<<"w ";
			else
				fout<<"  ";
	
			fout<<store[i]<<endl;
		}
	}
}
void write( ostream& fout )
{
	fout<<" Aux1 = " << Aux1;
#ifdef AUX2
	fout<<" Aux2 = " << Aux2;
	fout<<" Aux3 = " << Aux3;
#endif
#ifdef BYREF
	fout<<" lvusd  " <<last_var_used;
#endif //BYREF
	fout<<endl;
	for ( int i = 1; i<store_limit; i++ ) {
		if(store_used[i]!=0){
			fout <<i + Array_bot - 1 <<"\t";
			if((store_used[i]&store_read_unassigned)!=0)
				fout<<"u";
			else
				fout<<" ";
			if((store_used[i]&store_written)!=0)
				fout<<"w ";
			else
				fout<<"  ";
	
			fout<<store[i]<<endl;
		}
	}
}
//////////////////////////////////////////////////////////////////////

static retval print_buffer[print_buffer_size];
       int    print_buffer_position;

OPDEF(printEval)
{
#ifdef DEBUG
cout<<"printEval start "<<IP<<flush;
#endif

retval value = EVAL;

#ifdef DEBUG
cout<<"printEval end "<<IP<<endl;
#endif

if(print_buffer_position<print_buffer_size) {
	print_buffer[print_buffer_position++] = value;
	return print_buffer_position;
}
else
	Primitive_error |= print_overrun_error;
	return 0;
}

//inlined WBL 21 Feb 95
//void print_clear()
//{
//	print_buffer_position = 0;
//   //	memset(print_buffer,0,sizeof(print_buffer));
//}

void print_write(ostream& fout)
{
	fout<<"prtbuf=\t"<<print_buffer_position;
	for(int i = 0; i<print_buffer_position; i++) {
		fout<<" "<<print_buffer[i];
	}
}

BOOL print_equals(int length, retval* buffer)
{
	if (print_buffer_position != length)
		return FALSE;
	for(int i = 0; i<print_buffer_position; i++) {
		if(print_buffer[i] != buffer[i])
			return FALSE;
	}
	return TRUE;
}

float print_match(int length, retval* buffer, BOOL& exact)
{
	float score = 0.0;

	if (print_buffer_position == length) {
		if(memcmp(print_buffer,buffer,length*sizeof(retval))==0) {
			exact = TRUE;
			return print_match_max_score(length);
		}
	}
	score += print_match_length_score(length, print_buffer_position);

	for(int i = 0; i<length; i++) {
		for (int j = 0; j<print_buffer_position; j++) {
			if(print_buffer[j] == buffer[i]) {
				score += print_match_item_score(length);
				break; //leave j loop
			}
		}
	}

	float max_order_score = 0.0;
	for(int k = 0; k<length; k++) { 
	int   last_match_j = -1;
	float order_score  = 0.0;
	BOOL match = FALSE;
	for(int i = k; i<length; i++) {
		for (int j = 1+last_match_j; j<print_buffer_position; j++) {
			if(print_buffer[j] == buffer[i]) {
				last_match_j = j;
				order_score += print_match_order_score(length);
				match = TRUE;
				break; //leave loop j
			}
		}
	}
#ifdef DEBUG
	cout<<"k="<<k<<" order_score "<<order_score<<flush;
#endif
	if(order_score > max_order_score) max_order_score = order_score;
	if(max_order_score+0.00001 >= length*print_match_order_score(length)) {
#ifdef DEBUG
		cout<<"Reached print_match_order_score "<<max_order_score
		    <<" "<<k<<endl;
#endif
		break;
	}
	if( (print_buffer_position-k-1) * print_match_order_score(length)
	                                                   <= max_order_score){
#ifdef DEBUG
		cout<<"Cant exceed max_order_score "<<max_order_score
		    <<" "<<k<<endl;
#endif
		break;
        }
	if(match==FALSE) {
#ifdef DEBUG
		cout<<"No matches "<<k<<endl;
#endif
		break;
        }
	}//end loop k
	score += max_order_score;

	return score;
}

int randEval_seed = 1;

OPDEF(randEval) //assumes retval=float
{
	float range = EVAL;
	intrnd (randEval_seed);
	return range*(float(randEval_seed)/2147483647.0);
}

//LastEval
//////////////////////////////////////////////////////////////////////

void gp::AddF1 (int t, Function* f )
{
AddF(f);
addtofuncbag(t, f);
}

void gp::AddFbar1 (int exclude_tree, Function* f )
{
AddF(f);
for ( int i = 0; i < NUM_TREES; i++ )
	{if ( i != exclude_tree) addtofuncbag(i, f);};
}

void gp::AddF2 (int t1, int t2, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
}

void gp::AddF3 (int t1, int t2, int t3, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
}

void gp::AddF4 (int t1, int t2, int t3, int t4, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
}

void gp::AddF5 (int t1, int t2, int t3, int t4, int t5, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
}

void gp::AddF6 (int t1, int t2, int t3, int t4, int t5, int t6, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
}

void gp::AddF7 (int t1, int t2, int t3, 
		int t4, int t5, int t6, int t7, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
addtofuncbag(t7, f);
}

void gp::AddF8 (int t1, int t2, int t3, 
		int t4, int t5, int t6, int t7, int t8, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
addtofuncbag(t7, f);
addtofuncbag(t8, f);
}

void gp::AddF9 (int t1, int t2, int t3, 
		int t4, int t5, int t6, int t7, int t8, int t9, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
addtofuncbag(t7, f);
addtofuncbag(t8, f);
addtofuncbag(t9, f);
}

void gp::AddF11 (int t1, int t2, int t3,  int t4, int t5, int t6, int t7, 
                 int t8, int t9, int t10, int t11, Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
addtofuncbag(t7, f);
addtofuncbag(t8, f);
addtofuncbag(t9, f);
addtofuncbag(t10, f);
addtofuncbag(t11, f);
}

void gp::AddF13 (int t1, int t2, int t3,  int t4,  int t5,  int t6, int t7, 
                 int t8, int t9, int t10, int t11, int t12, int t13, 
		 Function* f )
{
AddF(f);
addtofuncbag(t1, f);
addtofuncbag(t2, f);
addtofuncbag(t3, f);
addtofuncbag(t4, f);
addtofuncbag(t5, f);
addtofuncbag(t6, f);
addtofuncbag(t7, f);
addtofuncbag(t8, f);
addtofuncbag(t9, f);
addtofuncbag(t10, f);
addtofuncbag(t11, f);
addtofuncbag(t12, f);
addtofuncbag(t13, f);
}


#ifdef ADF
void gp::AddFmain (Function* f )
{
AddFbar1(adf1, f);
}
#endif //ADF

void gp::AddFall (Function* f )
{
AddFbar1 (-1, f);
}


